"""
Unit test module for environment.py.
"""

# built-in imports
import os
import shutil
import sys
import tempfile
import unittest
import unittest.mock

# third-party imports

# local imports
import thermostatsupervisor
from thermostatsupervisor import emulator_config
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class EnvironmentTests(utc.UnitTest):
    """Test functions related to environment and env variables."""

    def setUp(self):
        super().setUp()
        util.log_msg.file_name = "unit_test.txt"

    def test_is_interactive_environment(self):
        """
        Verify is_interactive_environment().
        """
        return_val = env.is_interactive_environment()
        self.assertIsInstance(return_val, bool, "return value is not a boolean")

    def test_get_env_variable_with_default(self):
        """
        Test get_env_variable() with default value parameter.
        """
        # Test with missing variable and default value
        result = env.get_env_variable("NONEXISTENT_VAR", default="default_value")
        self.assertEqual(result["status"], util.NO_ERROR)
        self.assertEqual(result["value"], "default_value")
        self.assertEqual(result["key"], "NONEXISTENT_VAR")

        # Test with missing variable and no default (should fail as before)
        result = env.get_env_variable("NONEXISTENT_VAR")
        self.assertEqual(result["status"], util.ENVIRONMENT_ERROR)
        self.assertIsNone(result["value"])

        # Test with missing password variable and default value
        result = env.get_env_variable("MISSING_PASSWORD", default="secret123")
        self.assertEqual(result["status"], util.NO_ERROR)
        self.assertEqual(result["value"], "secret123")

        # Test with existing environment variable (default should be ignored)
        os.environ["TEST_EXISTING"] = "existing_value"
        try:
            result = env.get_env_variable("TEST_EXISTING", default="default_value")
            self.assertEqual(result["status"], util.NO_ERROR)
            self.assertEqual(result["value"], "existing_value")
        finally:
            del os.environ["TEST_EXISTING"]

    def test_get_flask_limiter_storage_uri(self):
        """
        Test get_flask_limiter_storage_uri() with and without environment variable.
        """
        # Test with no environment variable (should use default)
        result = env.get_flask_limiter_storage_uri()
        self.assertEqual(result, "memory://")

        # Test with environment variable set
        os.environ["FLASK_LIMITER_STORAGE_URI"] = "redis://localhost:6379"
        try:
            result = env.get_flask_limiter_storage_uri()
            self.assertEqual(result, "redis://localhost:6379")
        finally:
            del os.environ["FLASK_LIMITER_STORAGE_URI"]

    def test_get_env_variable_source_tracking(self):
        """
        Test that get_env_variable() properly tracks the source of variables.
        """
        # Test environment variable source
        os.environ["TEST_SOURCE_VAR"] = "from_env"
        try:
            result = env.get_env_variable("TEST_SOURCE_VAR")
            self.assertEqual(result["status"], util.NO_ERROR)
            self.assertEqual(result["value"], "from_env")
            self.assertEqual(result["source"], "environment_variable")
        finally:
            del os.environ["TEST_SOURCE_VAR"]

        # Test default value source
        result = env.get_env_variable("MISSING_VAR", default="default_val")
        self.assertEqual(result["status"], util.NO_ERROR)
        self.assertEqual(result["value"], "default_val")
        self.assertEqual(result["source"], "default")

    def test_get_env_variable_2fa_masking(self):
        """
        Test that 2FA keys are properly masked in get_env_variable().
        """
        # Test that 2FA keys are masked in logs
        os.environ["TEST_2FA"] = "123456"
        try:
            result = env.get_env_variable("TEST_2FA")
            self.assertEqual(result["status"], util.NO_ERROR)
            self.assertEqual(result["value"], "123456")
            # The actual masking is done in logging, not in the return value
        finally:
            del os.environ["TEST_2FA"]

    def test_get_env_variable(self):
        """
        Confirm get_env_variable() can retrieve values.
        """
        for env_key in ["GMAIL_USERNAME", "GMAIL_PASSWORD"]:
            buff = env.get_env_variable(env_key)
            print(
                f"env${env_key}="
                f"{[buff['value'], '(hidden)']['PASSWORD' in env_key]}"
            )
            self.assertEqual(buff["status"], util.NO_ERROR)
            self.assertGreater(len(buff["value"]), 0)

    def test_get_env_variable_from_file(self):
        """
        Test get_env_variable() with supervisor-env.txt file.
        """
        # Create temporary directory
        test_dir = tempfile.mkdtemp()
        original_cwd = os.getcwd()

        try:
            # Change to test directory
            os.chdir(test_dir)

            # Create supervisor-env.txt file
            env_content = """# Test environment file
TEST_KEY1=test_value1
TEST_KEY2=test_value2
# Comment line
TEST_PASSWORD=secret123

EMPTY_LINE_ABOVE=yes
"""
            with open("supervisor-env.txt", "w", encoding="utf-8") as f:
                f.write(env_content)

            # Test reading from file
            result = env.get_env_variable("TEST_KEY1")
            self.assertEqual(result["status"], util.NO_ERROR)
            self.assertEqual(result["value"], "test_value1")

            result = env.get_env_variable("TEST_KEY2")
            self.assertEqual(result["status"], util.NO_ERROR)
            self.assertEqual(result["value"], "test_value2")

            result = env.get_env_variable("TEST_PASSWORD")
            self.assertEqual(result["status"], util.NO_ERROR)
            self.assertEqual(result["value"], "secret123")

            # Test fallback to environment variable
            os.environ["FALLBACK_TEST"] = "fallback_value"
            result = env.get_env_variable("FALLBACK_TEST")
            self.assertEqual(result["status"], util.NO_ERROR)
            self.assertEqual(result["value"], "fallback_value")

            # Test missing variable
            result = env.get_env_variable("MISSING_VAR")
            self.assertEqual(result["status"], util.ENVIRONMENT_ERROR)

        finally:
            # Cleanup
            os.chdir(original_cwd)
            if "FALLBACK_TEST" in os.environ:
                del os.environ["FALLBACK_TEST"]
            shutil.rmtree(test_dir, ignore_errors=True)

    def test_read_supervisor_env_file_function(self):
        """
        Test the _read_supervisor_env_file() function directly.
        """
        # Create temporary directory
        test_dir = tempfile.mkdtemp()
        original_cwd = os.getcwd()

        try:
            # Change to test directory
            os.chdir(test_dir)

            # Test with no file
            result = env._read_supervisor_env_file()
            self.assertEqual(result, {})

            # Test with valid file
            env_content = """# Comment
KEY1=value1
KEY2=value with spaces
# Another comment

KEY3=value3
INVALID_LINE_NO_EQUALS
KEY4=value4=with=equals
"""
            with open("supervisor-env.txt", "w", encoding="utf-8") as f:
                f.write(env_content)

            result = env._read_supervisor_env_file()
            expected = {
                "KEY1": "value1",
                "KEY2": "value with spaces",
                "KEY3": "value3",
                "KEY4": "value4=with=equals",
            }
            self.assertEqual(result, expected)

        finally:
            # Cleanup
            os.chdir(original_cwd)
            shutil.rmtree(test_dir, ignore_errors=True)

    def test_load_all_env_variables(self):
        """
        Confirm all env variables can be loaded.
        """
        env.load_all_env_variables()
        print(f"env var dict={env.env_variables}")

    def test_get_local_ip(self):
        """
        Verify get_local_ip().
        """
        return_val = env.get_local_ip()
        self.assertTrue(
            isinstance(return_val, str),
            "get_local_ip() returned '%s' which is not a string",
        )
        self.assertTrue(
            7 <= len(return_val) <= 15,
            "get_local_ip() returned '%s' which is not between 7 and 15 chars",
        )

    def test_is_azure_environment(self):
        """
        Test is_azure_environment.
        """
        result = env.is_azure_environment()
        print(f"env.is_azure_environment()={result}")
        self.assertTrue(
            isinstance(result, bool),
            f"env.is_azure_environment() returned type {type(result)} expected bool",
        )

    def test_is_windows_environment(self):
        """
        Verify is_windows_environment() returns a bool.
        """
        return_val = env.is_windows_environment()
        self.assertIsInstance(return_val, bool, "return value is not a boolean")

    def test_is_raspberrypi_environment(self):
        """
        Test is_raspberrypi_environment.
        """
        result = env.is_raspberrypi_environment()
        print(f"env.is_raspberrypi_environment()={result}")
        self.assertTrue(
            isinstance(result, bool),
            "env.is_raspberrypi_environment() returned type "
            f"{type(result)} expected bool",
        )

    def test_get_python_version(self):
        """Verify get_python_version()."""
        major_version, minor_version = env.get_python_version()

        # verify major version
        min_major = int(env.MIN_PYTHON_MAJOR_VERSION)
        self.assertTrue(
            major_version >= min_major,
            f"python major version ({major_version}) is not gte "
            f"min required value ({min_major})",
        )

        # verify minor version
        min_minor = int(
            str(env.MIN_PYTHON_MAJOR_VERSION)[
                str(env.MIN_PYTHON_MAJOR_VERSION).find(".") + 1 :
            ]
        )
        self.assertTrue(
            minor_version >= min_minor,
            f"python minor version ({minor_version}) is not gte "
            f"min required value ({min_minor})",
        )

        # error checking invalid input parameter
        with self.assertRaises(TypeError):
            print("attempting to invalid input parameter type, expect exception...")
            env.get_python_version("3", 7)

        # no decimal point
        env.get_python_version(3, None)

        # min value exception
        with self.assertRaises(EnvironmentError):
            print("attempting to verify version gte 99.99, expect exception...")
            env.get_python_version(99, 99)

        print("test passed all checks")

    def test_dynamic_module_import(self):
        """
        Verify dynamic_module_import() runs without error

        TODO: this module results in a resourcewarning within unittest:
        sys:1: ResourceWarning: unclosed <socket.socket fd=628,
        family=AddressFamily.AF_INET, type=SocketKind.SOCK_DGRAM, proto=0,
        laddr=('0.0.0.0', 64963)>
        """

        # test successful case
        package_name = util.PACKAGE_NAME + "." + emulator_config.ALIAS
        pkg = env.dynamic_module_import(package_name)
        print(f"default thermostat returned package type {type(pkg)}")
        self.assertTrue(
            isinstance(pkg, object),
            f"dynamic_module_import() returned type({type(pkg)}),"
            f" expected an object",
        )
        del sys.modules[package_name]
        del pkg

        # test failing case
        with self.assertRaises(ImportError):
            print("attempting to open bogus package name, expect exception...")
            pkg = env.dynamic_module_import(util.PACKAGE_NAME + "." + "bogus")
            print(f"'bogus' module returned package type {type(pkg)}")
        print("test passed")

    def test_get_parent_path(self):
        """
        Verify get_parent_path().
        """
        return_val = env.get_parent_path(os.getcwd())
        self.assertTrue(
            isinstance(return_val, str),
            "get_parent_path() returned '%s' which is not a string",
        )

    def test_get_package_version(self):
        """
        Verify get_package_version().
        """
        pkg = thermostatsupervisor
        return_type = tuple
        return_val = env.get_package_version(pkg)
        self.assertTrue(
            isinstance(return_val, return_type),
            f"return_val = {return_val}, expected type "
            f"{return_type}, actual_type {type(return_val)}",
        )

        # check individual elements
        elements = [
            "major",
            "minor",
            "patch",
        ]
        return_type = int
        for element in elements:
            return_val = env.get_package_version(pkg, element)
            self.assertTrue(
                isinstance(return_val, return_type),
                f"element='{element}', return_val = {return_val},"
                " expected type "
                f"{return_type}, actual_type {type(return_val)}",
            )

    def test_show_package_version(self):
        """Verify show_package_version()."""
        env.show_package_version(thermostatsupervisor)

    def test_get_package_path(self):
        """Verify get_package_path()."""
        pkg = thermostatsupervisor
        return_val = env.get_package_path(pkg)
        self.assertTrue(
            isinstance(return_val, str),
            f"get_package_path() returned '{return_val}' which is not a string",
        )
        self.assertTrue(
            os.path.exists(return_val),
            f"get_package_path() returned '{return_val}' which does not exist",
        )
        self.assertTrue(
            return_val.endswith(".py"),
            f"get_package_path() returned '{return_val}' which is not a .py file",
        )

    def test_dynamic_module_import_no_reimport(self):
        """
        Verify dynamic_module_import() doesn't re-import local modules.
        """
        # Create a temporary module for testing
        test_dir = tempfile.mkdtemp()
        test_module_path = os.path.join(test_dir, "test_reuse_module.py")

        try:
            # Create a test module file
            with open(test_module_path, "w", encoding="utf-8") as f:
                f.write(
                    '''"""Test module for import reuse."""
__version__ = "1.0.0"
import_count = 0

def get_import_count():
    global import_count
    import_count += 1
    return import_count
'''
                )

            # First import should print warning
            mod1 = env.dynamic_module_import("test_reuse_module", path=test_dir)
            self.assertIsNotNone(mod1)
            self.assertTrue(hasattr(mod1, "get_import_count"))

            # Check that module is in sys.modules
            self.assertModuleIn("test_reuse_module")

            # Second import should NOT print warning
            initial_path_count = sys.path.count(test_dir)
            mod2 = env.dynamic_module_import("test_reuse_module", path=test_dir)
            final_path_count = sys.path.count(test_dir)

            # Verify the path wasn't added again
            self.assertEqual(
                initial_path_count,
                final_path_count,
                "Path was added to sys.path again during second import",
            )

            # Verify both imports return the same module object
            self.assertIs(mod1, mod2, "Second import returned different module object")

        finally:
            # Clean up
            if "test_reuse_module" in sys.modules:
                del sys.modules["test_reuse_module"]
            while test_dir in sys.path:
                sys.path.remove(test_dir)
            shutil.rmtree(test_dir, ignore_errors=True)

    def test_convert_to_absolute_path(self):
        """Verify convert_to_absolute_path()."""
        # Test with a valid relative path
        relative_path = "some/relative/path"
        absolute_path = env.convert_to_absolute_path(relative_path)
        self.assertTrue(
            os.path.isabs(absolute_path),
            f"convert_to_absolute_path() returned '{absolute_path}' which is not an "
            "absolute path",
        )

        # Test with an empty string
        relative_path = ""
        absolute_path = env.convert_to_absolute_path(relative_path)
        self.assertTrue(
            os.path.isabs(absolute_path),
            f"convert_to_absolute_path() returned '{absolute_path}' which is not an "
            "absolute path",
        )

        # Test with a None input
        with self.assertRaises(TypeError):
            env.convert_to_absolute_path(None)

        # Test with a non-string input
        with self.assertRaises(TypeError):
            env.convert_to_absolute_path(123)

    def test_set_env_variable(self):
        """Test set_env_variable() function."""
        # Test valid cases
        test_key = "TEST_SET_ENV_VAR"
        test_value = "test_value"

        try:
            env.set_env_variable(test_key, test_value)
            self.assertEqual(os.environ[test_key], test_value)

            # Test with integer value
            env.set_env_variable(test_key, 123)
            self.assertEqual(os.environ[test_key], "123")

            # Test with boolean value
            env.set_env_variable(test_key, True)
            self.assertEqual(os.environ[test_key], "True")

        finally:
            if test_key in os.environ:
                del os.environ[test_key]

        # Test error cases
        with self.assertRaises(AttributeError):
            env.set_env_variable(None, "value")

        with self.assertRaises(AttributeError):
            env.set_env_variable("key", None)

        with self.assertRaises(AttributeError):
            env.set_env_variable(123, "value")

    def test_add_package_to_path(self):
        """Test _add_package_to_path() function."""
        import sys

        # Store original sys.path
        original_path = sys.path.copy()

        try:
            # Test with None package (should do nothing)
            env._add_package_to_path(None)
            self.assertEqual(sys.path, original_path)

            # Test with valid package name
            test_pkg = "test_package"
            expected_path = env.get_parent_path(os.getcwd()) + "//" + test_pkg

            with unittest.mock.patch("builtins.print") as mock_print:
                env._add_package_to_path(test_pkg)

                # Verify package path was added to front of sys.path
                self.assertEqual(sys.path[0], expected_path)
                mock_print.assert_called_with(
                    f"adding package '{expected_path}' to path..."
                )

            # Test with verbose=True
            with unittest.mock.patch("builtins.print") as mock_print:
                env._add_package_to_path(test_pkg, verbose=True)

                # Should print twice - the adding message and the sys.path
                self.assertEqual(mock_print.call_count, 2)

        finally:
            # Restore original sys.path
            sys.path[:] = original_path


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Unit test module for supervise_flask_server.py.

Flask server tests currently do not work on Azure pipelines
because ports cannot be opened on shared pool.
"""

# built-in imports
import threading
import time
import unittest

# third party imports
from flask_wtf.csrf import CSRFProtect
import requests

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import flask_generic as flg

# thermostat_api is imported but not used to avoid a circular import
from thermostatsupervisor import (  # noqa F401, pylint: disable=unused-import.
    thermostat_api as api,
)
from thermostatsupervisor import supervisor_flask_server as sfs
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


@unittest.skipIf(
    env.is_azure_environment(), "this test not supported on Azure Pipelines"
)
@unittest.skipIf(
    not env.is_interactive_environment(),
    "this test hangs when run from the command line",
)
@unittest.skipIf(
    not utc.ENABLE_FLASK_INTEGRATION_TESTS, "flask integration tests are disabled"
)
class IntegrationTest(utc.UnitTest):
    """Test functions in supervisor_flask_server.py."""

    app = sfs.create_app()
    csrf = CSRFProtect(app)  # enable CSRF protection
    ip_ban = flg.initialize_ipban(app)  # hacker BlockListing agent
    flg.set_flask_cookie_config(app)
    flg.print_flask_config(app)

    def setUp(self):
        super().setUp()
        sfs.debug = False
        sfs.measurements = 10
        sfs.unit_test_mode = True
        util.log_msg.file_name = "unit_test.txt"
        if not env.is_azure_environment():
            # mock the argv list
            sfs.argv = utc.unit_test_argv
            print("starting supervise flask server thread...")
            self.flask_server = threading.Thread(
                target=sfs.app.run,
                args=("0.0.0.0", sfs.FLASK_PORT, False),
                kwargs=sfs.flask_kwargs,
            )
            self.flask_server.daemon = True  # make thread daemonic
            self.flask_server.start()
            print(f"thread alive status={self.flask_server.is_alive()}")
            print("Flask server setup is complete")
        else:
            print(
                "WARNING: flask server tests not currently supported on "
                "Azure pipelines, doing nothing"
            )

    def tearDown(self):
        if not env.is_azure_environment():
            print(f"thread alive status={self.flask_server.is_alive()}")
            if self.flask_server.daemon:
                print(
                    "flask server is daemon thread, "
                    "thread will terminate when main thread terminates"
                )
            else:
                print(
                    "WARNING: flask server is not daemon thread, "
                    "thread may still be active"
                )
        super().tearDown()

    def test_supervisor_flask_server(self):
        """
        Confirm Flask server returns valid data.

        This test requires a live thermostat connection to run the
        supervise routine on.
        """
        # grab supervise web page result and display
        flask_url = (
            sfs.FLASK_URL_PREFIX + env.get_local_ip() + ":" + str(sfs.FLASK_PORT)
        )

        # delay for page load and initial data posting
        wait_delay_sec = 10
        polling_interval_sec = 4
        while wait_delay_sec > 0:
            print(
                f"waiting {wait_delay_sec:.0f} seconds for initial "
                "supervisor page to be populated..."
            )
            wait_delay_sec -= polling_interval_sec
            time.sleep(polling_interval_sec)  # polling interval

        # grab web page and check response code
        print(f"grabbing web page results from: {flask_url}")
        results = requests.get(flask_url, timeout=util.HTTP_TIMEOUT)
        print(f"web page response code={results.status_code}")
        self.assertEqual(
            results.status_code,
            200,
            f"web page response was {results.status_code}, expected 200",
        )

        # check web page content vs. expectations
        print(f"web page contents: {results.content}")
        exp_substr = (
            f"<title>{utc.unit_test_argv[1]} thermostat zone "
            f"{utc.unit_test_argv[2]}, {utc.unit_test_argv[7]} "
            f"measurements</title>"
        )
        self.assertTrue(
            exp_substr in results.content.decode("utf-8"),
            f"did not find substring '{exp_substr}' in web page response",
        )


@unittest.skipIf(
    not utc.ENABLE_FLASK_INTEGRATION_TESTS, "flask integration tests are disabled"
)
class TestRunSupervise(utc.UnitTest):
    """Unit tests for run_supervise function."""

    def setUp(self):
        super().setUp()
        self.app = sfs.create_app()
        self.client = self.app.test_client()
        sfs.argv = utc.unit_test_argv

    def test_run_supervise_response(self):
        """Test the response of run_supervise function."""
        with self.app.test_request_context():
            response = sfs.index()
            self.assertEqual(response.status_code, 200)
            self.assertIn(
                f"<title>{utc.unit_test_argv[1]} thermostat zone "
                f"{utc.unit_test_argv[2]}, {utc.unit_test_argv[7]} "
                "measurements</title>",
                response.get_data(as_text=True),
            )

    def test_run_supervise_output(self):
        """Test the output of run_supervise function."""
        with self.app.test_request_context():
            response = sfs.index()
            # The response is a streaming response, consume the generator
            content_parts = list(response.response)
            content = "".join(content_parts)
            # At minimum, we should get the HTML title
            self.assertIn("<title>", content)
            # Note: <code> and <br> may not appear if subprocess doesn't
            # produce output. This is expected in test environments where
            # subprocess may not run properly


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Integration tests for weather functionality with thermostat reporting.
"""

# built-in imports
import inspect
import unittest
from unittest.mock import patch

# third-party imports

# local imports
from thermostatsupervisor import emulator_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import weather


class TestWeatherIntegration(unittest.TestCase):
    """Test weather integration with thermostat reporting."""

    def test_print_select_data_from_all_zones_signature(self):
        """Test that the function signature includes outdoor weather parameter."""

        sig = inspect.signature(tc.print_select_data_from_all_zones)

        # Check that all expected parameters are present
        expected_params = [
            "thermostat_type",
            "zone_lst",
            "ThermostatClass",
            "ThermostatZone",
            "display_wifi",
            "display_battery",
            "display_outdoor_weather",
        ]

        actual_params = list(sig.parameters.keys())
        self.assertEqual(actual_params, expected_params)

        # Check that new parameter has correct default
        outdoor_weather_param = sig.parameters["display_outdoor_weather"]
        self.assertTrue(outdoor_weather_param.default)

    def test_emulator_config_has_zip_code(self):
        """Test that emulator config includes zip code."""
        zip_code = emulator_config.supported_configs.get("zip_code")
        self.assertIsNotNone(zip_code)
        self.assertIsInstance(zip_code, str)
        self.assertEqual(zip_code, "55378")

    @patch("thermostatsupervisor.weather.get_outdoor_weather")
    @patch("thermostatsupervisor.weather.get_weather_api_key")
    def test_weather_data_integration(self, mock_get_api_key, mock_get_weather):
        """Test weather data integration in thermostat reporting."""
        # Mock weather data
        mock_get_api_key.return_value = None  # No API key, should use mock data
        mock_get_weather.return_value = {
            "outdoor_temp": 68.5,
            "outdoor_humidity": 45.0,
            "outdoor_conditions": "Sunny",
            "data_source": "mock",
        }

        # Test weather display formatting
        weather_data = mock_get_weather.return_value
        formatted = weather.format_weather_display(weather_data)
        expected = "outdoor: 68.5°F, 45%RH (Sunny)"
        self.assertEqual(formatted, expected)

    def test_config_integration(self):
        """Test that thermostat configurations properly include zip codes."""

        # Check that our modified configs have zip codes
        emulator_config_api = api.SUPPORTED_THERMOSTATS.get("emulator", {})
        self.assertIn("zip_code", emulator_config_api)
        self.assertEqual(emulator_config_api["zip_code"], "55378")

        honeywell_config_api = api.SUPPORTED_THERMOSTATS.get("honeywell", {})
        self.assertIn("zip_code", honeywell_config_api)
        self.assertEqual(honeywell_config_api["zip_code"], "55378")


if __name__ == "__main__":
    unittest.main()



"""
Integration test module for sht31_flask_server.py.

These tests require a working SHT31 endpoint and are controlled by the
ENABLE_FLASK_INTEGRATION_TESTS flag.
"""

# built-in imports
import unittest

# third party imports

# local imports
# thermostat_api is imported but not used to avoid a circular import
from thermostatsupervisor import environment as env
from thermostatsupervisor import flask_generic as flg
from thermostatsupervisor import (
    thermostat_api as api,
)  # noqa F401, pylint: disable=unused-import.
from thermostatsupervisor import sht31
from thermostatsupervisor import sht31_config
from thermostatsupervisor import sht31_flask_server as sht31_fs
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


@unittest.skipIf(not utc.ENABLE_SHT31_TESTS, "sht31 tests are disabled")
@unittest.skipIf(
    env.is_azure_environment(), "this test not supported on Azure Pipelines"
)
@unittest.skipIf(
    not utc.ENABLE_FLASK_INTEGRATION_TESTS, "flask integration tests are disabled"
)
class IntegrationTest(utc.UnitTest):
    """Test functions in sht31_flask_server.py."""

    # sht31 flask server is automatically spawned in sht31
    # Thermostat class if unit test zone is being used.

    def test_sht31_flask_server_all_pages(self):
        """
        Confirm all pages return data from Flask server.
        """
        # do not test these pages
        no_test_list = ["i2c_recovery", "reset"]

        # no server outptu for these pages
        no_server_output_list = []
        no_key_check_list = ["print_block_list", "clear_block_list"]
        # loopback does not work so use local sht31 zone if testing
        # on the local net.  If not, use the DNS name.
        zone = sht31_config.get_preferred_zone()
        # Define expected keys for each test case
        expected_keys = {
            "production": "measurements",
            "unit_test": "measurements",
            "diag": "raw_binary",
            "clear_diag": "raw_binary",
            "enable_heater": "raw_binary",
            "disable_heater": "raw_binary",
            "soft_reset": "raw_binary",
            "i2c_detect": "i2c_detect",
            "i2c_detect_0": "i2c_detect",
            "i2c_detect_1": "i2c_detect",
            "i2c_logic_levels": "i2c_logic_levels",
            "i2c_bus_health": "i2c_bus_health",
            "i2c_recovery": "i2c_recovery",
            "reset": "message",
        }

        for test_case in sht31_config.flask_folder:
            if test_case in no_test_list:
                print(f"test_case={test_case}: bypassing this test case")
                continue

            print(f"test_case={test_case}")
            Thermostat = sht31.ThermostatClass(
                zone, path=sht31_config.flask_folder[test_case]
            )
            print("printing thermostat meta data:")
            return_data = Thermostat.print_all_thermostat_metadata(zone)

            # validate return type was returned
            if test_case in no_server_output_list:
                self.assertTrue(
                    isinstance(return_data, type(None)),
                    f"return data for test case {test_case} is not NoneType, "
                    f"return type: {type(return_data)}",
                )
            else:
                self.assertTrue(
                    isinstance(return_data, dict),
                    f"return data for test case {test_case} is not a dictionary, "
                    f"return type: {type(return_data)}",
                )

            # validate key as proof of correct return page
            if test_case not in no_key_check_list:
                expected_key = expected_keys.get(test_case, "bogus")
                self.assertTrue(
                    expected_key in return_data,
                    f"test_case '{test_case}': key '{expected_key}' "
                    f"was not found in return data: {return_data}",
                )

    def test_sht31_flask_server(self):
        """
        Confirm Flask server returns valid data.
        """
        measurements_bckup = sht31_config.MEASUREMENTS
        try:
            for sht31_config.measurements in [1, 10, 100, 1000]:
                msg = ["measurement", "measurements"][sht31_config.MEASUREMENTS > 1]
                print(
                    f"\ntesting SHT31 flask server with "
                    f"{sht31_config.MEASUREMENTS} {msg}..."
                )
                self.validate_flask_server()
        finally:
            sht31_config.measurements = measurements_bckup

    def validate_flask_server(self):
        """
        Launch SHT31 Flask server and validate data.
        """
        print("creating thermostat object...")
        Thermostat = sht31.ThermostatClass(sht31_config.UNIT_TEST_ZONE)
        print("printing thermostat meta data:")
        Thermostat.print_all_thermostat_metadata(sht31_config.UNIT_TEST_ZONE)

        # create mock runtime args
        api.uip = api.UserInputs(utc.unit_test_sht31)

        # create Zone object
        Zone = sht31.ThermostatZone(Thermostat)

        # update runtime overrides
        Zone.update_runtime_parameters()

        print("current thermostat settings...")
        print(f"switch position: {Zone.get_system_switch_position()}")
        print(f"heat mode={Zone.is_heat_mode()}")
        print(f"cool mode={Zone.is_cool_mode()}")
        print(f"temporary hold minutes={Zone.get_temporary_hold_until_time()}")
        meta_data = Thermostat.get_all_metadata(sht31_config.UNIT_TEST_ZONE, retry=True)
        print(f"thermostat meta data={meta_data}")
        print(
            f"thermostat display temp="
            f"{util.temp_value_with_units(Zone.get_display_temp())}"
        )

        # verify measurements
        self.assertEqual(
            meta_data["measurements"],
            sht31_config.MEASUREMENTS,
            f"measurements: actual={meta_data['measurements']}, "
            f"expected={sht31_config.MEASUREMENTS}",
        )

        # verify metadata
        test_cases = {
            "get_display_temp": {"min_val": 80, "max_val": 120},
            "get_is_humidity_supported": {"min_val": True, "max_val": True},
            "get_display_humidity": {"min_val": 49, "max_val": 51},
        }
        for param, limits in test_cases.items():
            return_val = getattr(Zone, param)()
            print(f"'{param}'={return_val}")
            min_val = limits["min_val"]
            max_val = limits["max_val"]
            self.assertTrue(
                min_val <= return_val <= max_val,
                f"'{param}'={return_val}, not between {min_val} and {max_val}",
            )
        # cleanup
        del Zone
        del Thermostat


@unittest.skipIf(not utc.ENABLE_SHT31_TESTS, "sht31 tests are disabled")
@unittest.skipIf(
    not utc.ENABLE_FLASK_INTEGRATION_TESTS, "flask integration tests are disabled"
)
class TestSht31FlaskClientAzure(utc.UnitTest):
    """
    Azure-compatible tests for SHT31 Flask server using test client.

    These tests use Flask's test_client() which doesn't require network access,
    making them suitable for Azure Pipelines environments.
    """

    def setUp(self):
        super().setUp()
        self.app = sht31_fs.create_app()
        self.client = self.app.test_client()
        self.app.config["TESTING"] = True
        # Initialize IP ban for testing
        self.ip_ban = flg.initialize_ipban(self.app)

    def test_sht31_flask_server_endpoints_response(self):
        """Test that SHT31 Flask server endpoints return valid responses."""
        # Define endpoints that should return 200 status
        test_endpoints = [
            ("/data", "production"),
            ("/unit", "unit_test"),
            ("/diag", "diag"),
            ("/clear_diag", "clear_diag"),
            ("/enable_heater", "enable_heater"),
            ("/disable_heater", "disable_heater"),
            ("/soft_reset", "soft_reset"),
            ("/i2c_detect", "i2c_detect"),
            ("/i2c_detect/0", "i2c_detect_0"),
            ("/i2c_detect/1", "i2c_detect_1"),
            ("/i2c_logic_levels", "i2c_logic_levels"),
            ("/i2c_bus_health", "i2c_bus_health"),
            ("/print_block_list", "print_block_list"),
            ("/clear_block_list", "clear_block_list"),
            # Note: skipping '/reset' and '/i2c_recovery' for side effects
        ]

        for endpoint, test_name in test_endpoints:
            with self.subTest(endpoint=endpoint, test_name=test_name):
                try:
                    response = self.client.get(endpoint)
                    # Check that we get a valid response
                    self.assertIn(
                        response.status_code,
                        [200, 400, 404, 500],
                        f"Endpoint {endpoint} returned unexpected "
                        f"status {response.status_code}",
                    )

                    # For successful responses, check that we get JSON data
                    if response.status_code == 200:
                        self.assertTrue(
                            response.is_json or response.mimetype == "application/json",
                            f"Endpoint {endpoint} should return JSON data",
                        )
                        data = response.get_json()
                        self.assertIsInstance(
                            data, dict, f"Endpoint {endpoint} should return dict data"
                        )
                except Exception as e:
                    # In test environment, some endpoints may fail due to
                    # missing hardware. This is acceptable as we're testing
                    # the Flask routing, not hardware
                    print(
                        f"Warning: Endpoint {endpoint} failed with {e}, "
                        f"this may be expected in test environment"
                    )

    def test_sht31_flask_server_unit_test_endpoint(self):
        """Test the unit_test endpoint specifically."""
        try:
            response = self.client.get("/unit")

            # Should get a valid response (404 is OK if endpoint doesn't exist)
            self.assertIn(response.status_code, [200, 400, 404, 500])

            if response.status_code == 200:
                # Should return JSON data
                self.assertTrue(
                    response.is_json or response.mimetype == "application/json"
                )
                data = response.get_json()
                self.assertIsInstance(data, dict)
                # Should contain measurements key for unit test endpoint
                if "measurements" in data:
                    self.assertIsInstance(data["measurements"], int)
        except (FileNotFoundError, OSError) as e:
            # Expected in test environments where system utilities like
            # iwconfig aren't available
            print(f"Expected error in test environment: {e}")
            # This is actually a positive result - it means the routing worked
            # and we got to the correct endpoint code

    def test_sht31_flask_server_bad_routes(self):
        """Test that bad routes return proper 404 responses."""
        # Define non-existent endpoints that should return 404
        bad_endpoints = [
            "/nonexistent",
            "/bad_route",
            "/invalid_endpoint",
            "/not_found",
            "/fake_api",
            "/i2c_detect/999",  # Invalid bus number
            "/enable_heater/extra",  # Extra path component
        ]

        for endpoint in bad_endpoints:
            with self.subTest(endpoint=endpoint):
                response = self.client.get(endpoint)
                # Should get 404 for non-existent routes
                self.assertEqual(
                    response.status_code,
                    404,
                    f"Endpoint {endpoint} should return 404, "
                    f"got {response.status_code}",
                )

    def test_sht31_flask_server_ipban_registry(self):
        """Test IP ban registry functionality."""
        # Test the print_block_list endpoint
        response = self.client.get("/print_block_list")
        self.assertIn(
            response.status_code,
            [200, 500],  # 500 is OK in test environment
            f"print_block_list endpoint returned {response.status_code}",
        )

        if response.status_code == 200:
            # Should return JSON data
            self.assertTrue(
                response.is_json or response.mimetype == "application/json",
                "print_block_list should return JSON data",
            )
            data = response.get_json()
            self.assertIsInstance(
                data, dict, "print_block_list should return dict data"
            )
            # Block list should be a dictionary (even if empty)
            self.assertIsInstance(
                data, dict, "IP ban block list should be a dictionary"
            )

    def test_sht31_flask_server_clear_ipban_registry(self):
        """Test clearing the IP ban registry."""
        # Test the clear_block_list endpoint
        response = self.client.get("/clear_block_list")
        self.assertIn(
            response.status_code,
            [200, 500],  # 500 is OK in test environment
            f"clear_block_list endpoint returned {response.status_code}",
        )

        if response.status_code == 200:
            # Should return JSON data
            self.assertTrue(
                response.is_json or response.mimetype == "application/json",
                "clear_block_list should return JSON data",
            )
            data = response.get_json()
            self.assertIsInstance(
                data, dict, "clear_block_list should return dict data"
            )
            # After clearing, block list should be empty
            self.assertEqual(len(data), 0, "Block list should be empty after clearing")

    def test_sht31_flask_server_ipban_recovery_workflow(self):
        """Test the complete IP ban and recovery workflow."""

        # Step 1: Clear the block list to ensure clean state
        clear_response = self.client.get("/clear_block_list")
        if clear_response.status_code == 200:
            cleared_data = clear_response.get_json()
            self.assertIsInstance(cleared_data, dict)
            self.assertEqual(
                len(cleared_data), 0, "Block list should be empty after clearing"
            )

            # Step 2: Check initial block list state
            print_response = self.client.get("/print_block_list")
            if print_response.status_code == 200:
                initial_data = print_response.get_json()
                self.assertIsInstance(initial_data, dict)
                self.assertEqual(
                    len(initial_data), 0, "Block list should be empty initially"
                )

                # Step 3: Test bad route triggering IP ban
                # Make requests to bad routes to trigger IP ban
                # Use ipban_ban_count to determine how many requests needed
                ban_count_needed = flg.ipban_ban_count
                print(f"Making {ban_count_needed} bad requests to trigger IP ban")

                # Test with known nuisance patterns that should trigger bans
                # These patterns are typically loaded by default in flask-ipban
                test_bad_routes = [
                    "/wp-admin",  # Common vulnerability scan target
                    "/admin",  # Another common target
                    "/.env",  # Environment file scanning
                    "/phpMyAdmin",  # Database admin tool scanning
                ]

                for i in range(ban_count_needed):
                    # Use both custom and known bad routes
                    bad_route = test_bad_routes[i % len(test_bad_routes)]
                    if i >= len(test_bad_routes):
                        bad_route = f"/nonexistent_route_{i}"

                    bad_response = self.client.get(bad_route)
                    self.assertEqual(
                        bad_response.status_code,
                        404,
                        f"Bad route {bad_route} should return 404",
                    )

                # Step 4: Check if IP was added to block list after bad requests
                check_response = self.client.get("/print_block_list")
                if check_response.status_code == 200:
                    blocked_data = check_response.get_json()
                    self.assertIsInstance(blocked_data, dict)
                    print(
                        f"Block list after {ban_count_needed} bad requests: "
                        f"{blocked_data}"
                    )

                    # Step 5: Since automatic IP banning may not work in test
                    # environment, manually add an IP to test the clear functionality
                    # This ensures we test the core IP ban registry management
                    with self.app.test_request_context():
                        # Manually add an IP to the ban list for testing purposes
                        self.ip_ban.add("192.168.1.100", url="/test_ban")
                        manual_blocked_data = self.ip_ban.get_block_list()
                        print(
                            f"Block list after manual addition: "
                            f"{manual_blocked_data}"
                        )

                        # Verify that the IP was added
                        self.assertGreater(
                            len(manual_blocked_data),
                            0,
                            "Block list should contain entries after manual addition",
                        )

                    # Step 6: Test the clear functionality
                    final_clear_response = self.client.get("/clear_block_list")
                    if final_clear_response.status_code == 200:
                        final_cleared_data = final_clear_response.get_json()
                        self.assertIsInstance(final_cleared_data, dict)
                        self.assertEqual(
                            len(final_cleared_data),
                            0,
                            "Block list should be empty after final clearing",
                        )

                        # Step 7: Verify block list is empty after clearing
                        verify_response = self.client.get("/print_block_list")
                        if verify_response.status_code == 200:
                            verify_data = verify_response.get_json()
                            self.assertIsInstance(verify_data, dict)
                            self.assertEqual(
                                len(verify_data),
                                0,
                                "Block list should remain empty after verification",
                            )


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""Blink Camera."""

# built-in imports
import asyncio
import pprint
import sys
import time
import traceback
from typing import Union
from aiohttp import ClientSession

# third party imports

# local imports
from thermostatsupervisor import blink_config
from thermostatsupervisor import environment as env
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import utilities as util

# Blink library
BLINK_DEBUG = True  # debug uses local blink repo instead of pkg
if BLINK_DEBUG and not env.is_azure_environment():
    pkg = "blinkpy.blinkpy"
    mod_path = "..\\blinkpy"
    if env.is_interactive_environment():
        mod_path = "..\\" + mod_path
    blinkpy = env.dynamic_module_import("blinkpy.blinkpy", mod_path, pkg)
    auth = env.dynamic_module_import("blinkpy.auth", mod_path, pkg)
else:
    from blinkpy import auth  # noqa E402, from path / site packages
    from blinkpy import blinkpy  # noqa E402, from path / site packages

# Import blinkpy exceptions for proper error handling
try:
    from blinkpy.auth import LoginError, UnauthorizedError
    from aiohttp.client_exceptions import ClientConnectionError, ContentTypeError
except ImportError:
    # Fallback for older versions or missing exceptions
    class LoginError(Exception):
        """Login error fallback."""

        pass

    class UnauthorizedError(Exception):
        """Unauthorized error fallback."""

        pass

    class ClientConnectionError(Exception):
        """Client connection error fallback."""

        pass

    class ContentTypeError(Exception):
        """Content type error fallback."""

        pass


class ThermostatClass(blinkpy.Blink, tc.ThermostatCommon):
    """Blink Camera thermostat functions."""

    def __init__(self, zone, verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            zone(str):  zone of thermostat.
            verbose(bool): debug flag.
        """
        # Blink server auth credentials from env vars
        self.BL_UNAME_KEY = "BLINK_USERNAME"
        self.BL_PASSWORD_KEY = "BLINK_PASSWORD"
        self.BL_2FA_KEY = "BLINK_2FA"

        # Get username
        uname_result = env.get_env_variable(
            self.BL_UNAME_KEY,
            default="<" + self.BL_UNAME_KEY + api.KEY_MISSING_SUFFIX
        )
        self.bl_uname = uname_result["value"]

        # Get password
        pwd_result = env.get_env_variable(
            self.BL_PASSWORD_KEY,
            default="<" + self.BL_PASSWORD_KEY + api.KEY_MISSING_SUFFIX
        )
        self.bl_pwd = pwd_result["value"]

        # Get 2FA with detailed logging
        twofa_result = env.get_env_variable(
            self.BL_2FA_KEY,
            default="<" + self.BL_2FA_KEY + api.KEY_MISSING_SUFFIX
        )
        self.bl_2fa = twofa_result["value"]
        self._log_2fa_source(twofa_result)

        self.auth_dict = {"username": self.bl_uname, "password": self.bl_pwd}
        self.verbose = verbose
        self.zone_number = int(zone)

        # connect to Blink server and authenticate
        self.args = None
        self.thermostat_type = None
        self.blink = None
        if env.get_package_version(blinkpy) >= (0, 22, 0):
            asyncio.run(self.async_auth_start())
        else:
            self.auth_start()

        # get cameras
        self.camera_metadata = {}
        self.get_cameras()

        # configure zone info
        self.zone_name = self.get_zone_name()
        self.device_id = None  # initialize
        self.device_id = self.get_target_zone_id(self.zone_number)
        self.serial_number = None  # will be populated when unit is queried.

    def _log_2fa_source(self, twofa_result):
        """
        Log the source of the 2FA code with appropriate masking.

        inputs:
            twofa_result(dict): result from get_env_variable with source info
        returns:
            None
        """
        source = twofa_result.get("source", "unknown")
        value = twofa_result.get("value", "")

        # Create source message
        if source == "supervisor-env.txt":
            source_msg = "using stored 2FA from supervisor-env.txt"
        elif source == "environment_variable":
            source_msg = "using stored 2FA from environment variable"
        elif source == "default":
            source_msg = "using default 2FA value (missing)"
        else:
            source_msg = f"using 2FA from {source}"

        # Mask or show 2FA based on debug mode
        debug_enabled = getattr(util.log_msg, "debug", False)
        if debug_enabled:
            # Show actual 2FA in debug mode
            twofa_display = f"2FA code: {value}"
        else:
            # Mask 2FA in non-debug mode
            if value and not value.startswith("<"):
                twofa_display = "2FA code: ******"
            else:
                twofa_display = f"2FA code: {value}"

        # Log the information
        util.log_msg(
            f"Blink zone {self.zone_number}: {source_msg}, {twofa_display}",
            mode=util.STDOUT_LOG + util.DATA_LOG,
        )

    def _handle_auth_retry(self, attempt, max_retries, retry_delay, error):
        """Handle authentication retry logic."""
        if attempt < max_retries - 1:
            if self.verbose:
                print(
                    f"Authentication attempt {attempt + 1} failed: {str(error)}. "
                    f"Retrying in {retry_delay} seconds..."
                )
            time.sleep(retry_delay)
            return retry_delay * 2  # exponential backoff
        else:
            # Final attempt failed
            error_msg = self._format_auth_error(error, "sync")
            banner = "*" * len(error_msg)
            print(banner)
            print(error_msg)
            print(banner)
            sys.exit(1)

    def _handle_setup_retry(self, attempt, max_retries, retry_delay):
        """Handle setup post-verification retry logic."""
        if attempt < max_retries - 1:
            if self.verbose:
                print(
                    f"Post-verification setup failed, retrying in "
                    f"{retry_delay} seconds... "
                    f"(attempt {attempt + 1})"
                )
            time.sleep(retry_delay)
            return True  # Continue retry loop
        else:
            raise RuntimeError(
                "Blink post-verification setup failed after retries. "
                "Camera list may not be available."
            )

    def _attempt_authentication(self):
        """Attempt single authentication process."""
        self.blink = blinkpy.Blink()
        if self.blink is None:
            raise RuntimeError(
                "ERROR: Blink object failed to instantiate "
                f"for zone {self.zone_number}"
            )

        self.blink.auth = auth.Auth(self.auth_dict, no_prompt=True)
        self.blink.start()

        # Send 2FA key with proper error checking
        auth_success = self.blink.auth.send_auth_key(self.blink, self.bl_2fa)
        if not auth_success:
            raise ValueError(
                "2FA verification failed. Please check your verification code."
            )

        # Check if setup_post_verify succeeds with retry
        setup_success = self.blink.setup_post_verify()
        return setup_success

    def auth_start(self):
        """
        blinkpy < 0.22.0-compatible start with improved error handling
        """
        self._setup_auth_parameters()
        self._execute_auth_with_retry()

    def _setup_auth_parameters(self):
        """Setup authentication parameters."""
        self.args = [self.bl_uname, self.bl_pwd]
        self.thermostat_type = blink_config.ALIAS

    def _execute_auth_with_retry(self):
        """Execute authentication with retry logic."""
        max_retries = 3
        retry_delay = 5  # seconds
        expected_exceptions = (
            AttributeError,
            ValueError,
            KeyError,
            LoginError,
            UnauthorizedError,
            ClientConnectionError,
            ContentTypeError,
        )

        for attempt in range(max_retries):
            try:
                setup_success = self._attempt_authentication()
                if not setup_success:
                    if self._handle_setup_retry(attempt, max_retries, retry_delay):
                        continue
                break

            except expected_exceptions as e:
                retry_delay = self._handle_auth_retry(
                    attempt, max_retries, retry_delay, e
                )
                continue
            except Exception as e:
                self._handle_unexpected_error(e)

    async def _execute_async_auth_with_retry(self, session):
        """Execute async authentication with retry logic."""
        max_retries = 3
        retry_delay = 5  # seconds
        expected_exceptions = (
            AttributeError,
            ValueError,
            KeyError,
            LoginError,
            UnauthorizedError,
            ClientConnectionError,
            ContentTypeError,
        )

        for attempt in range(max_retries):
            try:
                setup_success = await self._attempt_async_authentication(session)
                if not setup_success:
                    if await self._handle_async_setup_retry(
                        attempt, max_retries, retry_delay
                    ):
                        continue
                break

            except expected_exceptions as e:
                retry_delay = await self._handle_async_auth_retry(
                    attempt, max_retries, retry_delay, e
                )
                continue
            except Exception as e:
                self._handle_unexpected_error(e)

    def _handle_unexpected_error(self, error):
        """Handle unexpected errors during authentication."""
        print(traceback.format_exc())

        # Check for specific error patterns
        error_str = str(error)
        if (
            "homescreen" in error_str.lower()
            or "token refresh" in error_str.lower()
        ):
            error_msg = (
                f"ERROR: Blink authentication failed for zone "
                f"{self.zone_number}. The server rejected the request after "
                f"token refresh, likely due to an invalid or expired 2FA code. "
                f"2FA codes from authenticator apps expire after 30-60 seconds. "
                f"Please update your {self.BL_2FA_KEY} environment variable or "
                f"supervisor-env.txt file with a fresh code from your "
                f"authenticator app and restart the application. "
                f"Error: {error_str}"
            )
        else:
            error_msg = (
                f"ERROR: Unexpected error during Blink authentication for zone "
                f"{self.zone_number}: {error_str}"
            )

        banner = "*" * len(error_msg)
        print(banner)
        print(error_msg)
        print(banner)
        sys.exit(1)

    async def _handle_async_auth_retry(self, attempt, max_retries, retry_delay, error):
        """Handle async authentication retry logic."""
        if attempt < max_retries - 1:
            if self.verbose:
                print(
                    f"Authentication attempt {attempt + 1} failed: "
                    f"{str(error)}. Retrying in {retry_delay} seconds..."
                )
            await asyncio.sleep(retry_delay)
            return retry_delay * 2  # exponential backoff
        else:
            # Final attempt failed
            error_msg = self._format_auth_error(error, "async")
            banner = "*" * len(error_msg)
            print(banner)
            print(error_msg)
            print(banner)
            sys.exit(1)

    async def _handle_async_setup_retry(self, attempt, max_retries, retry_delay):
        """Handle async setup post-verification retry logic."""
        if attempt < max_retries - 1:
            if self.verbose:
                print(
                    f"Post-verification setup failed, retrying in "
                    f"{retry_delay} seconds... (attempt {attempt + 1})"
                )
            await asyncio.sleep(retry_delay)
            return True  # Continue retry loop
        else:
            raise RuntimeError(
                "Blink post-verification setup failed after "
                "retries. Camera list may not be available."
            )

    async def _attempt_async_authentication(self, session):
        """Attempt single async authentication process."""
        self.blink = blinkpy.Blink(session=session)
        if self.blink is None:
            raise RuntimeError(
                "ERROR: Blink object failed to instantiate "
                f"for zone {self.zone_number}"
            )

        self.blink.auth = auth.Auth(self.auth_dict, no_prompt=True, session=session)
        await self.blink.start()

        # Send 2FA key with proper error checking
        auth_success = await self.blink.auth.send_auth_key(self.blink, self.bl_2fa)
        if not auth_success:
            raise ValueError(
                "2FA verification failed. " "Please check your verification code."
            )

        # Check if setup_post_verify succeeds with retry
        setup_success = await self.blink.setup_post_verify()
        return setup_success

    async def async_auth_start(self):
        """
        blinkpy 0.22.0 introducted async start, this is the compatible
        auth_start function with improved error handling and retry logic.
        """
        async with ClientSession() as session:
            self._setup_auth_parameters()
            await self._execute_async_auth_with_retry(session)

    def _format_auth_error(self, error, auth_type="sync"):
        """
        Format authentication error messages with specific guidance.

        inputs:
            error: The exception that occurred
            auth_type: "sync" or "async" to indicate which auth method failed
        returns:
            (str): Formatted error message with troubleshooting guidance
        """
        error_type = type(error).__name__
        error_str = str(error)

        error_handlers = {
            LoginError: self._format_login_error,
            UnauthorizedError: self._format_unauthorized_error,
            ClientConnectionError: self._format_connection_error,
            ContentTypeError: self._format_content_error,
            ValueError: self._format_value_error,
            AttributeError: self._format_attribute_error,
        }

        handler = error_handlers.get(type(error))
        if handler:
            return handler(error_str, auth_type)

        return self._format_generic_error(error_type, error_str, auth_type)

    def _format_login_error(self, error_str, auth_type):
        """Format login error message."""
        return (
            f"ERROR: Blink login failed for zone {self.zone_number} "
            f"({auth_type} mode). Please check your username and password. "
            f"The Blink server may also be down or experiencing issues. "
            f"Error: {error_str}"
        )

    def _format_unauthorized_error(self, error_str, auth_type):
        """Format unauthorized error message."""
        return (
            f"ERROR: Blink authorization failed for zone {self.zone_number} "
            f"({auth_type} mode). Your account may be locked or credentials "
            f"may be invalid. Error: {error_str}"
        )

    def _format_connection_error(self, error_str, auth_type):
        """Format connection error message."""
        return (
            f"ERROR: Network connection to Blink servers failed for zone "
            f"{self.zone_number} ({auth_type} mode). Please check your "
            f"internet connection and try again. Error: {error_str}"
        )

    def _format_content_error(self, error_str, auth_type):
        """Format content type error message."""
        return (
            f"ERROR: Received invalid response from Blink servers for zone "
            f"{self.zone_number} ({auth_type} mode). The server may be "
            f"experiencing issues. Error: {error_str}"
        )

    def _format_value_error(self, error_str, auth_type):
        """Format value error message."""
        if (
            "2FA verification failed" in error_str
            or "Invalid Verification Code" in error_str
        ):
            return (
                f"ERROR: Invalid 2FA verification code for zone "
                f"{self.zone_number} ({auth_type} mode). The 2FA code may "
                f"be expired or incorrect. 2FA codes from authenticator apps "
                f"expire after 30-60 seconds. Please update your "
                f"{self.BL_2FA_KEY} environment variable or "
                f"supervisor-env.txt file with a fresh code from your "
                f"authenticator app and restart the application. "
                f"Error: {error_str}"
            )
        return self._format_generic_error("ValueError", error_str, auth_type)

    def _format_attribute_error(self, error_str, auth_type):
        """Format attribute error message."""
        return (
            f"ERROR: Blink authentication state error for zone "
            f"{self.zone_number} ({auth_type} mode). This may occur if the "
            f"initial login failed. Please verify your credentials and try "
            f"again. Error: {error_str}"
        )

    def _format_generic_error(self, error_type, error_str, auth_type):
        """Format generic error message."""
        return (
            f"ERROR: Blink authentication failed for zone "
            f"{self.zone_number} ({auth_type} mode). Error type: "
            f"{error_type}. Error details: {error_str}"
        )

    def get_zone_name(self):
        """
        Return the name associated with the zone number from metadata dict.

        inputs:
            None
        returns:
            (str) zone name
        """
        return blink_config.metadata[self.zone_number]["zone_name"]

    def get_target_zone_id(self, zone=0):
        """
        Return the target zone ID.

        inputs:
            zone(int): zone number.
        returns:
            (obj): Blink object
        """
        # return the target zone object
        return zone

    def get_cameras(self):
        """
        Get the blink cameras
        """
        table_length = 20
        if self.verbose:
            print("blink camera inventory:")
            print("-" * table_length)

        if not self.blink.cameras:
            if self.verbose:
                print("WARNING: No cameras found in blink.cameras")
                print("This may indicate authentication or setup issues")
            return

        for name, camera in self.blink.cameras.items():
            if self.verbose:
                print(name)
                print(camera.attributes)
            self.camera_metadata[name] = camera.attributes
        if self.verbose:
            print(f"Total cameras found: {len(self.blink.cameras)}")
            print("-" * table_length)

    def get_all_metadata(self, zone=None, retry=False):
        """Get all thermostat meta data for device_id from local API.

        inputs:
            zone(): specified zone
            retry(bool): if True will retry with extended retry mechanism
        returns:
            (dict): dictionary of meta data.
        """
        return self.get_metadata(zone, retry=retry)

    def _handle_empty_camera_list_async(self, zone_name):
        """Handle empty camera list for async version of blinkpy."""
        available_cameras = (
            list(self.blink.cameras.keys()) if self.blink.cameras else []
        )
        error_msg = (
            f"Camera list is empty when searching for camera "
            f"'{zone_name}'. Available cameras: "
            f"{available_cameras}. This may indicate "
            f"authentication tokens have expired. Please restart "
            f"the application to re-authenticate."
        )
        raise ValueError(error_msg)

    def _handle_token_refresh_failure(self, zone_name, refresh_error):
        """Handle token refresh failure with helpful error message."""
        if self.verbose:
            print(f"Token refresh failed: {str(refresh_error)}")

        available_cameras = (
            list(self.blink.cameras.keys()) if self.blink.cameras else []
        )
        error_msg = (
            f"Camera list is empty when searching for "
            f"camera '{zone_name}'. Available cameras: "
            f"{available_cameras}. Authentication token "
            f"refresh failed: {str(refresh_error)}. Please "
            f"restart the application to re-authenticate."
        )
        raise ValueError(error_msg)

    def _attempt_sync_token_refresh(self, zone_name):
        """Attempt to refresh authentication token for sync version."""
        try:
            self._perform_token_refresh()
            self._refresh_camera_data()
        except Exception as refresh_error:
            self._handle_token_refresh_failure(zone_name, refresh_error)

    def _perform_token_refresh(self):
        """Perform the actual token refresh."""
        if self.verbose:
            print("Attempting to refresh authentication token...")
        self.blink.auth.refresh_token()

    def _refresh_camera_data(self):
        """Refresh camera data after token refresh."""
        if hasattr(self.blink, "refresh"):
            self.blink.refresh()
        elif hasattr(self.blink, "setup_camera_list"):
            self.blink.setup_camera_list()

        # Update our local camera metadata cache
        self.get_cameras()

    def _refresh_camera_list_if_empty(self, zone_name):
        """Refresh camera list if it's empty."""
        if self.blink.cameras != {}:
            return

        self._log_camera_refresh_attempt(zone_name)

        if not hasattr(self.blink.auth, "refresh_token"):
            return

        try:
            self._attempt_camera_refresh(zone_name)
            self._validate_camera_refresh_success(zone_name)
        except Exception as e:
            self._handle_camera_refresh_error(e, zone_name)

    def _log_camera_refresh_attempt(self, zone_name):
        """Log camera refresh attempt."""
        if self.verbose:
            print(
                f"Camera list is empty, attempting to refresh authentication "
                f"and camera list for zone {zone_name}"
            )

    def _attempt_camera_refresh(self, zone_name):
        """Attempt to refresh camera list based on blinkpy version."""
        if env.get_package_version(blinkpy) >= (0, 22, 0):
            self._handle_empty_camera_list_async(zone_name)
        else:
            self._attempt_sync_token_refresh(zone_name)

    def _validate_camera_refresh_success(self, zone_name):
        """Validate that camera refresh was successful."""
        if self.blink.cameras == {}:
            error_msg = (
                f"Camera list is still empty after refresh attempt "
                f"for camera '{zone_name}'. This may indicate "
                f"authentication failed or no cameras are available. "
                f"Please check your Blink credentials, 2FA code, and "
                f"ensure cameras are online in the Blink app."
            )
            raise ValueError(error_msg)

    def _handle_camera_refresh_error(self, error, zone_name):
        """Handle camera refresh errors."""
        if "Camera list is empty when searching" in str(error):
            raise
        else:
            raise ValueError(
                f"Camera list is empty when searching for camera "
                f"'{zone_name}'. Failed to refresh camera list: {str(error)}"
            )

    def _find_camera_by_name(self, zone_name, parameter):
        """Find camera by name and return its attributes or specific parameter."""
        for name, camera in self.blink.cameras.items():
            if name == zone_name:
                if self.verbose:
                    print(f"found camera {name}: {camera.attributes}")
                if parameter is None:
                    return camera.attributes
                else:
                    return camera.attributes[parameter]

        # Camera not found - provide helpful error
        available_cameras = list(self.blink.cameras.keys())
        error_msg = (
            f"Camera zone '{zone_name}' was not found. "
            f"Available cameras: {available_cameras}. "
            f"Please check the zone name in blink_config.py matches "
            f"your Blink app."
        )
        raise ValueError(error_msg)

    def get_metadata(self, zone=None, trait=None, parameter=None, retry=False):
        """Get thermostat meta data for device_id from local API.

        inputs:
            zone(): specified zone
            trait(str): trait or parent key, if None will assume a non-nested
            dict
            parameter(str): target parameter, if None will return all.
            retry(bool): if True will retry with extended retry mechanism
        returns:
            (dict): dictionary of meta data.
        """
        del trait  # unused on blink

        def _get_metadata_internal():
            zone_name = blink_config.metadata[self.zone_number]["zone_name"]
            self._refresh_camera_list_if_empty(zone_name)
            return self._find_camera_by_name(zone_name, parameter)

        if retry:
            # Use standardized extended retry mechanism
            return util.execute_with_extended_retries(
                func=_get_metadata_internal,
                thermostat_type=getattr(self, "thermostat_type", "Blink"),
                zone_name=str(getattr(self, "zone_name", zone)),
                number_of_retries=5,
                initial_retry_delay_sec=60,
                exception_types=(
                    ValueError,
                    KeyError,
                    AttributeError,
                    ConnectionError,
                    TimeoutError,
                ),
                email_notification=None,  # Blink doesn't import email_notification
            )
        else:
            # Single attempt without retry
            return _get_metadata_internal()

    def print_all_thermostat_metadata(self, zone):
        """Print all metadata for zone to the screen.

        inputs:
            zone(int): specified zone, if None will print all zones.
        returns:
            None, prints result to screen
        """
        self.exec_print_all_thermostat_metadata(self.get_all_metadata, [zone])


class ThermostatZone(tc.ThermostatCommonZone):
    """
    KumoCloud single zone on local network.

    Class needs to be updated for multi-zone support.
    """

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Zone constructor.

        inputs:
            Thermostat(obj): Thermostat class instance.
            verbose(bool): debug flag.
        """
        # construct the superclass, requires auth setup first
        super().__init__()

        # runtime parameter defaults
        self.poll_time_sec = 10 * 60  # default to 10 minutes
        self.connection_time_sec = 8 * 60 * 60  # default to 8 hours

        # server data cache expiration parameters to mitigate spam detection
        self.fetch_interval_sec = 60  # age of server data before refresh (seconds)
        self.last_fetch_time = time.time() - 2 * self.fetch_interval_sec
        self.last_printed_refresh_time = None  # track last printed cache message time

        # switch config for this thermostat
        self.system_switch_position[tc.ThermostatCommonZone.COOL_MODE] = "cool"
        self.system_switch_position[tc.ThermostatCommonZone.HEAT_MODE] = "heat"
        self.system_switch_position[tc.ThermostatCommonZone.OFF_MODE] = "off"
        self.system_switch_position[tc.ThermostatCommonZone.DRY_MODE] = "dry"
        self.system_switch_position[tc.ThermostatCommonZone.AUTO_MODE] = "auto"
        self.system_switch_position[tc.ThermostatCommonZone.FAN_MODE] = "vent"

        # zone info
        self.verbose = verbose
        self.thermostat_type = blink_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.Thermostat = Thermostat_obj
        self.zone_number = Thermostat_obj.zone_number
        self.zone_name = self.get_zone_name()
        self.zone_metadata = Thermostat_obj.get_metadata(zone=self.zone_number)

    def get_display_temp(self) -> float:  # used
        """
        Refresh the cached zone information and return Indoor Temp in °F.

        inputs:
            None
        returns:
            (float): indoor temp in °F.
        """
        # Refresh zone metadata if needed (respects cache timeout)
        self.refresh_zone_info()

        raw_temp = self.zone_metadata.get(blink_config.API_TEMPF_MEAN)
        if isinstance(raw_temp, (str, float, int)):
            raw_temp = float(raw_temp)
        elif isinstance(raw_temp, type(None)):
            raw_temp = float(util.BOGUS_INT)
        return raw_temp

    def get_zone_name(self) -> str:
        """
        Return the name associated with the zone number from metadata dict.

        inputs:
            None
        returns:
            (str) zone name
        """
        return blink_config.metadata[self.zone_number]["zone_name"]

    def get_display_humidity(self) -> Union[float, None]:
        """
        Refresh the cached zone information and return IndoorHumidity.

        inputs:
            None
        returns:
            (float, None): indoor humidity in %RH, None if not supported.
        """
        return None  # not available

    def get_is_humidity_supported(self) -> bool:
        """Return humidity sensor status."""
        return self.get_display_humidity() is not None

    def is_heat_mode(self) -> int:
        """Return the heat mode."""
        return 0  # not applicable

    def is_cool_mode(self) -> int:
        """Return the cool mode."""
        return 0  # not applicable

    def is_dry_mode(self) -> int:
        """Return the dry mode."""
        return 0  # not applicable

    def is_auto_mode(self) -> int:
        """Return the auto mode."""
        return 0  # not applicable

    def is_eco_mode(self) -> int:
        """Return the auto mode."""
        return 0  # not applicable

    def is_fan_mode(self) -> int:
        """Return the fan mode."""
        return 0  # not applicable

    def is_off_mode(self) -> int:
        """Return the off mode."""
        return 1  # always off

    def is_heating(self) -> int:
        """Return 1 if actively heating, else 0."""
        return 0  # not applicable

    def is_cooling(self) -> int:
        """Return 1 if actively cooling, else 0."""
        return 0  # not applicable

    def is_drying(self) -> int:
        """Return 1 if drying relay is active, else 0."""
        return 0  # not applicable

    def is_auto(self) -> int:
        """Return 1 if auto relay is active, else 0."""
        return 0  # not applicable

    def is_eco(self) -> int:
        """Return 1 if eco relay is active, else 0."""
        return 0  # not applicable

    def is_fanning(self) -> int:
        """Return 1 if fan relay is active, else 0."""
        return 0  # not applicable

    def is_power_on(self) -> int:
        """Return 1 if power relay is active, else 0."""
        return 1  # always on

    def is_fan_on(self) -> int:
        """Return 1 if fan relay is active, else 0."""
        return 0  # not applicable

    def is_defrosting(self) -> int:
        """Return 1 if defrosting is active, else 0."""
        return 0  # not applicable

    def is_standby(self) -> int:
        """Return 1 if standby is active, else 0."""
        return 0  # not applicable

    def get_system_switch_position(self) -> int:
        """Return the thermostat mode.

        inputs:
            None
        returns:
            (int): thermostat mode, see tc.system_switch_position for details.
        """
        return self.system_switch_position[self.OFF_MODE]

    def get_wifi_strength(self) -> float:  # noqa R0201
        """Return the wifi signal strength in dBm."""
        # Refresh zone metadata if needed (respects cache timeout)
        self.refresh_zone_info()

        raw_wifi = self.zone_metadata.get(blink_config.API_WIFI_STRENGTH)
        if isinstance(raw_wifi, (str, float, int)):
            return float(raw_wifi)
        else:
            return float(util.BOGUS_INT)

    def get_wifi_status(self) -> bool:  # noqa R0201
        """Return the wifi connection status."""
        raw_wifi = self.get_wifi_strength()
        if isinstance(raw_wifi, (float, int)):
            return raw_wifi >= util.MIN_WIFI_DBM
        else:
            return False

    def get_battery_voltage(self) -> float:  # noqa R0201
        """Return the battery voltage in volts."""
        # Refresh zone metadata if needed (respects cache timeout)
        self.refresh_zone_info()

        raw_voltage = self.zone_metadata.get(blink_config.API_BATTERY_VOLTAGE)
        if isinstance(raw_voltage, (str, float, int)):
            return float(raw_voltage) / 100.0
        else:
            return float(util.BOGUS_INT)

    def get_battery_status(self) -> bool:  # noqa R0201
        """Return the battery status."""
        # Refresh zone metadata if needed (respects cache timeout)
        self.refresh_zone_info()

        raw_status = self.zone_metadata.get(blink_config.API_BATTERY_STATUS)
        if isinstance(raw_status, str):
            raw_status = raw_status == "ok"
        return raw_status

    def refresh_zone_info(self, force_refresh=False) -> None:
        """
        Refresh zone metadata from blink server with spam mitigation.

        This method overrides the base class to properly refresh blink camera
        data while respecting cache intervals to prevent server spam detection.

        inputs:
            force_refresh(bool): if True, ignore expiration timer and refresh
        returns:
            None, updates self.zone_metadata with fresh data from server
        """
        now_time = time.time()
        # refresh if past expiration date or force_refresh option
        if force_refresh or (
            now_time >= (self.last_fetch_time + self.fetch_interval_sec)
        ):
            if self.verbose:
                util.log_msg(
                    f"Refreshing zone metadata for {self.zone_name} "
                    f"(last refresh: {now_time - self.last_fetch_time:.1f}s ago)",
                    mode=util.STDOUT_LOG,
                    func_name=1,
                )

            # Get fresh metadata from blink server
            try:
                self.zone_metadata = self.Thermostat.get_metadata(zone=self.zone_number)
                self.last_fetch_time = now_time
                if self.verbose:
                    util.log_msg(
                        f"Zone metadata refreshed successfully for {self.zone_name}",
                        mode=util.STDOUT_LOG,
                        func_name=1,
                    )
            except Exception as e:
                if self.verbose:
                    util.log_msg(
                        f"Failed to refresh zone metadata for {self.zone_name}: {e}",
                        mode=util.STDOUT_LOG,
                        func_name=1,
                    )
                # Don't update last_fetch_time on failure to retry sooner
                raise
        else:
            if self.verbose:
                time_until_refresh = (
                    self.last_fetch_time + self.fetch_interval_sec - now_time
                )
                # Only log if refresh time has changed significantly from last print
                rounded_refresh_time = round(time_until_refresh)
                if (self.last_printed_refresh_time is None or
                        abs(rounded_refresh_time -
                            self.last_printed_refresh_time) >= 1):
                    util.log_msg(
                        f"Using cached data for {self.zone_name} "
                        f"(refresh in {time_until_refresh:.1f}s)",
                        mode=util.STDOUT_LOG,
                        func_name=1,
                    )
                    self.last_printed_refresh_time = rounded_refresh_time


if __name__ == "__main__":
    # verify environment
    env.get_python_version()
    env.show_package_version(blinkpy)

    # get zone override
    api.uip = api.UserInputs(argv_list=None, thermostat_type=blink_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name, api.input_flds.zone)

    _, Zone = tc.thermostat_basic_checkout(
        blink_config.ALIAS, zone_number, ThermostatClass, ThermostatZone
    )

    # this code is rem'd out because it will trigger blink server spam detectors.
    # tc.print_select_data_from_all_zones(
    #     blink_config.ALIAS,
    #     blink_config.get_available_zones(),
    #     ThermostatClass,
    #     ThermostatZone,
    #     display_wifi=True,
    #     display_battery=True,
    # )

    # measure thermostat response time
    if blink_config.check_response_time:
        MEASUREMENTS = 30
        meas_data = Zone.measure_thermostat_repeatability(
            MEASUREMENTS,
            func=Zone.pyhtcc.get_zones_info,
            measure_response_time=True,
        )
        ppp = pprint.PrettyPrinter(indent=4)
        ppp.pprint(meas_data)



"""
emulator thermostat config file.
"""

ALIAS = "emulator"

# constants
MAX_HEAT_SETPOINT = 66.0  # float
MIN_COOL_SETPOINT = 78.0  # float
STARTING_MODE = "OFF_MODE"  # thermostat set mode when emulator starts
STARTING_TEMP = 72.0  # starting temperature when emulator starts
NORMAL_TEMP_VARIATION = 16.0  # reported value variation +/- this value
STARTING_HUMIDITY = 45.0  # starting humidity when emulator starts
NORMAL_HUMIDITY_VARIATION = 3.0  # reported val variation +/- this val

# all environment variables specific to this thermostat type
env_variables = {}

# min required env variables on all runs
required_env_variables = {}

# supported thermostat configs
supported_configs = {
    "module": "emulator",
    "type": 0,
    "zones": [0, 1],
    "modes": [
        "OFF_MODE",
        "HEAT_MODE",
        "COOL_MODE",
        "DRY_MODE",
        "AUTO_MODE",
        "UNKNOWN_MODE",
    ],
    "zip_code": "55378",  # Zip code for outdoor weather data
}


def get_available_zones():
    """
    Return list of available zones.

    for this thermostat type, available zones is all zones.

    inputs:
        None.
    returns:
        (list) available zones.
    """
    return supported_configs["zones"]


default_zone = supported_configs["zones"][0]
default_zone_name = ALIAS + "_" + str(default_zone)

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "19",  # poll time in sec
    "359",  # reconnect time in sec
    "3",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
]

# flag to check thermostat response time during basic checkout
check_response_time = False



"""
Unit tests for environment.py package version detection.
"""

import unittest
from unittest.mock import MagicMock, patch

from thermostatsupervisor import environment as env
from tests import unit_test_common as utc


class TestPackageVersionDetection(utc.UnitTest):
    """Test package version detection functionality."""

    def setUp(self):
        """Set up test fixtures."""
        super().setUp()
        self.print_test_name()

    def test_get_package_version_with_version_attribute(self):
        """Test version detection when module has __version__ attribute."""
        # Create a mock module with __version__
        mock_module = MagicMock()
        mock_module.__name__ = "test_module"
        mock_module.__version__ = "1.2.3"

        version = env.get_package_version(mock_module)
        self.assertEqual(version, (1, 2, 3))

    def test_get_package_version_without_version_attribute(self):
        """Test version detection fallback using importlib.metadata."""
        # Create a mock module without __version__
        mock_module = MagicMock()
        mock_module.__name__ = "blinkpy"
        del mock_module.__version__  # Remove __version__ attribute

        # Mock importlib.metadata to return a known version
        with patch("importlib.metadata.version") as mock_version:
            mock_version.return_value = "0.23.0"
            version = env.get_package_version(mock_module)
            self.assertEqual(version, (0, 23, 0))
            mock_version.assert_called_once_with("blinkpy")

    def test_get_package_version_fallback_to_zero(self):
        """Test version detection fallback to 0.0.0 when all methods fail."""
        # Create a mock module without __version__
        mock_module = MagicMock()
        mock_module.__name__ = "nonexistent_module"
        del mock_module.__version__

        # Mock importlib.metadata to raise PackageNotFoundError
        with patch("importlib.metadata.version") as mock_version:
            import importlib.metadata

            mock_version.side_effect = importlib.metadata.PackageNotFoundError()
            version = env.get_package_version(mock_module)
            self.assertEqual(version, (0, 0, 0))

    def test_get_package_version_with_dev_suffix(self):
        """Test version detection strips dev suffixes correctly."""
        mock_module = MagicMock()
        mock_module.__name__ = "test_module"
        mock_module.__version__ = "1.2.3.dev0+abc123"

        version = env.get_package_version(mock_module)
        self.assertEqual(version, (1, 2, 3))

    def test_get_package_version_element_selection(self):
        """Test individual element selection from version tuple."""
        mock_module = MagicMock()
        mock_module.__name__ = "test_module"
        mock_module.__version__ = "2.5.7"

        # Test major version
        major = env.get_package_version(mock_module, element="major")
        self.assertEqual(major, 2)

        # Test minor version
        minor = env.get_package_version(mock_module, element="minor")
        self.assertEqual(minor, 5)

        # Test patch version
        patch = env.get_package_version(mock_module, element="patch")
        self.assertEqual(patch, 7)

        # Test numeric element selection
        major_num = env.get_package_version(mock_module, element=0)
        self.assertEqual(major_num, 2)

    def test_blink_version_detection_integration(self):
        """Integration test for blinkpy version detection."""
        # This test requires blinkpy to be installed
        try:
            import blinkpy

            version = env.get_package_version(blinkpy)

            # Version should be a tuple of three integers
            self.assertIsInstance(version, tuple)
            self.assertEqual(len(version), 3)
            self.assertTrue(all(isinstance(v, int) for v in version))

            # Version should be >= (0, 22, 0) for modern blinkpy
            self.assertGreaterEqual(
                version,
                (0, 22, 0),
                "blinkpy version should be >= 0.22.0 for async authentication",
            )

        except ImportError:
            self.skipTest("blinkpy not installed")


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
Unit tests for the SSL certificate download script.
"""

import unittest
from unittest.mock import patch

from download_ssl_certificates import parse_servers, main


class TestDownloadSSLCertificatesScript(unittest.TestCase):
    """Test the SSL certificate download script functionality."""

    def test_parse_servers_json_format(self):
        """Test parsing servers from JSON format."""
        # Test JSON array with dict objects
        servers_json = (
            '[{"hostname": "example.com", "port": 443}, '
            '{"hostname": "test.com", "port": 8443}]'
        )
        result = parse_servers(servers_json)
        expected = [("example.com", 443), ("test.com", 8443)]
        self.assertEqual(result, expected)

        # Test JSON array with string objects
        servers_json = '["example.com:443", "test.com:8443", "default.com"]'
        result = parse_servers(servers_json)
        expected = [("example.com", 443), ("test.com", 8443), ("default.com", 443)]
        self.assertEqual(result, expected)

    def test_parse_servers_comma_separated(self):
        """Test parsing servers from comma-separated format."""
        servers_str = "example.com:443,test.com:8443,default.com"
        result = parse_servers(servers_str)
        expected = [("example.com", 443), ("test.com", 8443), ("default.com", 443)]
        self.assertEqual(result, expected)

    def test_parse_servers_invalid_port(self):
        """Test parsing servers with invalid port number."""
        servers_str = "example.com:invalid"
        with self.assertRaises(ValueError) as context:
            parse_servers(servers_str)
        self.assertIn("Invalid port number", str(context.exception))

    def test_parse_servers_empty_input(self):
        """Test parsing empty server input."""
        result = parse_servers("")
        self.assertEqual(result, [])

        result = parse_servers("[]")
        self.assertEqual(result, [])

    @patch("download_ssl_certificates.download_and_import_ssl_certificates")
    @patch(
        "sys.argv", ["download_ssl_certificates.py", "example.com:443,test.com:8443"]
    )
    def test_main_download_and_import(self, mock_download_import):
        """Test main function with download and import."""
        mock_download_import.return_value = True

        result = main()
        self.assertEqual(result, 0)
        mock_download_import.assert_called_once_with(
            [("example.com", 443), ("test.com", 8443)]
        )

    @patch("thermostatsupervisor.ssl_certificate.download_ssl_certificate")
    @patch(
        "sys.argv",
        ["download_ssl_certificates.py", "example.com:443", "--download-only"],
    )
    def test_main_download_only(self, mock_download):
        """Test main function with download-only flag."""
        mock_download.return_value = "test_cert_path"

        result = main()
        self.assertEqual(result, 0)
        mock_download.assert_called_once_with("example.com", 443)

    @patch("download_ssl_certificates.download_and_import_ssl_certificates")
    @patch("sys.argv", ["download_ssl_certificates.py", "invalid_format"])
    def test_main_with_invalid_servers(self, mock_download_import):
        """Test main function with invalid server format."""
        # When parse_servers returns empty list, but it actually
        # converts "invalid_format" to a valid server
        # Let's use a truly empty server list case
        mock_download_import.return_value = True
        result = main()
        # This will actually succeed since "invalid_format" becomes a hostname
        self.assertEqual(result, 0)

    @patch("download_ssl_certificates.download_and_import_ssl_certificates")
    @patch("sys.argv", ["download_ssl_certificates.py", "example.com:443"])
    def test_main_with_failure(self, mock_download_import):
        """Test main function when certificate processing fails."""
        mock_download_import.return_value = False

        result = main()
        self.assertEqual(result, 1)  # Should return error code

    @patch("thermostatsupervisor.ssl_certificate.download_ssl_certificate")
    @patch(
        "sys.argv",
        [
            "download_ssl_certificates.py",
            "example.com:443,test.com:8443",
            "--download-only",
        ],
    )
    def test_main_download_only_partial_failure(self, mock_download):
        """Test main function with download-only and partial failure."""
        # First download succeeds, second fails
        mock_download.side_effect = ["test_cert_path", RuntimeError("Download failed")]

        result = main()
        # Should return error code due to partial failure
        self.assertEqual(result, 1)
        self.assertEqual(mock_download.call_count, 2)


if __name__ == "__main__":
    unittest.main()



"""
WSGI entry point for supervisor Flask server production deployment.

This module provides the WSGI application object that can be used by
production WSGI servers like Gunicorn.

Example usage with Gunicorn:
    gunicorn --config gunicorn.supervisor.conf.py \\
        thermostatsupervisor.supervisor_wsgi:application
"""

# local imports
from thermostatsupervisor import supervisor_flask_server as sfs
from thermostatsupervisor import utilities as util

# Enable logging to STDERR for production
util.log_stdout_to_stderr = True

# Create the WSGI application
application = sfs.create_app()



"""
SSL Certificate management utilities.

This module provides functionality to generate and manage self-signed SSL
certificates for Flask servers.
"""

# built-in imports
import pathlib
import platform
import subprocess
import time
from typing import Tuple, Optional, List

# local imports
from thermostatsupervisor import utilities as util


def get_ssl_cert_directory() -> pathlib.Path:
    """Get the directory where SSL certificates should be stored.

    Returns:
        pathlib.Path: Path to SSL certificate directory
    """
    # Create ssl directory in the project root
    project_root = pathlib.Path(__file__).parent.parent
    ssl_dir = project_root / "ssl"
    ssl_dir.mkdir(exist_ok=True)
    return ssl_dir


def generate_self_signed_certificate(
    cert_file: str = "server.crt",
    key_file: str = "server.key",
    days: int = 365,
    common_name: str = "localhost",
) -> Tuple[pathlib.Path, pathlib.Path]:
    """Generate a self-signed SSL certificate using OpenSSL.

    Args:
        cert_file: Name of the certificate file (default: server.crt)
        key_file: Name of the private key file (default: server.key)
        days: Number of days the certificate is valid (default: 365)
        common_name: Common name for the certificate (default: localhost)

    Returns:
        Tuple of (cert_path, key_path) as pathlib.Path objects

    Raises:
        RuntimeError: If OpenSSL command fails or is not available
    """
    ssl_dir = get_ssl_cert_directory()
    cert_path = ssl_dir / cert_file
    key_path = ssl_dir / key_file

    # Check if certificates already exist and are recent
    if cert_path.exists() and key_path.exists():
        # Check if certificate is still valid (created within last 30 days)
        cert_age_days = (time.time() - cert_path.stat().st_mtime) / (24 * 3600)
        if cert_age_days < (days - 30):  # Regenerate 30 days before expiry
            util.log_msg(
                f"Using existing SSL certificate: {cert_path}", mode=util.DEBUG_LOG
            )
            return cert_path, key_path

    util.log_msg(f"Generating new SSL certificate: {cert_path}", mode=util.STDOUT_LOG)

    # Build OpenSSL command with platform-specific configuration
    openssl_cmd = [
        "openssl",
        "req",
        "-x509",
        "-newkey",
        "rsa:4096",
        "-nodes",
        "-out",
        str(cert_path),
        "-keyout",
        str(key_path),
        "-days",
        str(days),
        "-subj",
        f"/C=US/ST=State/L=City/O=Organization/CN={common_name}",
    ]

    # Windows-specific configuration to avoid config file issues
    if platform.system().lower() == "windows":
        # Add -config flag to bypass default config file loading on Windows
        # This prevents the "Unable to load config info" error on Windows
        openssl_cmd.extend(["-config", "nul"])

    try:
        # Run OpenSSL command
        subprocess.run(
            openssl_cmd, capture_output=True, text=True, check=True, timeout=30
        )

        util.log_msg("SSL certificate generated successfully", mode=util.STDOUT_LOG)

        # Verify files were created
        if not cert_path.exists() or not key_path.exists():
            raise RuntimeError("Certificate files were not created")

        # Set proper permissions (readable by owner only)
        cert_path.chmod(0o600)
        key_path.chmod(0o600)

        return cert_path, key_path

    except subprocess.CalledProcessError as e:
        error_msg = f"OpenSSL command failed: {e.stderr}"
        util.log_msg(error_msg, mode=util.STDERR_LOG)
        raise RuntimeError(error_msg) from e

    except subprocess.TimeoutExpired as e:
        error_msg = "OpenSSL command timed out"
        util.log_msg(error_msg, mode=util.STDERR_LOG)
        raise RuntimeError(error_msg) from e

    except FileNotFoundError as e:
        error_msg = (
            "OpenSSL not found. Please install OpenSSL to generate " "SSL certificates"
        )
        util.log_msg(error_msg, mode=util.STDERR_LOG)
        raise RuntimeError(error_msg) from e


def get_ssl_context(
    cert_file: str = "server.crt",
    key_file: str = "server.key",
    fallback_to_adhoc: bool = True,
) -> Optional[str]:
    """Get SSL context for Flask application.

    Args:
        cert_file: Name of the certificate file
        key_file: Name of the private key file
        fallback_to_adhoc: Whether to fallback to 'adhoc' if cert generation fails

    Returns:
        SSL context tuple (cert_path, key_path) or 'adhoc' or None
    """
    try:
        cert_path, key_path = generate_self_signed_certificate(
            cert_file=cert_file, key_file=key_file
        )
        return (str(cert_path), str(key_path))

    except RuntimeError as e:
        util.log_msg(f"Failed to generate SSL certificate: {e}", mode=util.STDERR_LOG)

        if fallback_to_adhoc:
            util.log_msg(
                "Falling back to Flask 'adhoc' SSL certificate", mode=util.STDOUT_LOG
            )
            return "adhoc"
        else:
            return None


def validate_ssl_certificate(cert_path: pathlib.Path) -> bool:
    """Validate an SSL certificate file.

    Args:
        cert_path: Path to the certificate file

    Returns:
        True if certificate is valid, False otherwise
    """
    if not cert_path.exists():
        return False

    try:
        # Build OpenSSL validation command
        openssl_cmd = ["openssl", "x509", "-in", str(cert_path), "-noout", "-text"]

        # Windows-specific configuration to avoid config file issues
        if platform.system().lower() == "windows":
            # Add -config flag to bypass default config file loading on Windows
            openssl_cmd.extend(["-config", "nul"])

        # Use OpenSSL to verify the certificate
        subprocess.run(
            openssl_cmd,
            capture_output=True,
            text=True,
            check=True,
            timeout=10,
        )
        return True

    except (
        subprocess.CalledProcessError,
        subprocess.TimeoutExpired,
        FileNotFoundError,
    ):
        return False


def download_ssl_certificate(hostname: str, port: int = 443) -> pathlib.Path:
    """Download SSL certificate from a remote server.

    Args:
        hostname: The hostname/IP address of the server
        port: The port number (default: 443)

    Returns:
        Path to the downloaded certificate file

    Raises:
        RuntimeError: If certificate download fails
    """
    ssl_dir = get_ssl_cert_directory()
    cert_filename = f"{hostname}_{port}.crt"
    cert_path = ssl_dir / cert_filename

    util.log_msg(
        f"Downloading SSL certificate from {hostname}:{port}", mode=util.STDOUT_LOG
    )

    try:
        # Use openssl command to get the certificate
        openssl_cmd = [
            "openssl",
            "s_client",
            "-connect",
            f"{hostname}:{port}",
            "-servername",
            hostname,
            "-showcerts",
        ]

        # Windows-specific configuration to avoid config file issues
        if platform.system().lower() == "windows":
            # Add -config flag to bypass default config file loading on Windows
            openssl_cmd.extend(["-config", "nul"])

        # Run openssl command
        result = subprocess.run(
            openssl_cmd, input="", capture_output=True, text=True, timeout=30
        )

        if result.returncode != 0:
            raise RuntimeError(f"OpenSSL command failed: {result.stderr}")

        # Extract the certificate from the output
        output = result.stdout
        cert_start = output.find("-----BEGIN CERTIFICATE-----")
        cert_end = output.find("-----END CERTIFICATE-----") + len(
            "-----END CERTIFICATE-----"
        )

        if cert_start == -1 or cert_end == -1:
            raise RuntimeError("Could not find certificate in OpenSSL output")

        cert_pem = output[cert_start:cert_end]

        # Write certificate to file
        with open(cert_path, "w", encoding="utf-8") as f:
            f.write(cert_pem)

        # Set proper permissions
        cert_path.chmod(0o644)

        util.log_msg(f"SSL certificate downloaded to {cert_path}", mode=util.STDOUT_LOG)
        return cert_path

    except subprocess.TimeoutExpired:
        error_msg = f"Timeout downloading SSL certificate from {hostname}:{port}"
        util.log_msg(error_msg, mode=util.STDERR_LOG)
        raise RuntimeError(error_msg)
    except Exception as e:
        error_msg = f"Failed to download SSL certificate from {hostname}:{port}: {e}"
        util.log_msg(error_msg, mode=util.STDERR_LOG)
        raise RuntimeError(error_msg) from e


def import_ssl_certificate_to_system(cert_path: pathlib.Path) -> bool:
    """Import SSL certificate to system trust store.

    Args:
        cert_path: Path to the certificate file

    Returns:
        True if import was successful, False otherwise
    """
    if not cert_path.exists():
        util.log_msg(f"Certificate file not found: {cert_path}", mode=util.STDERR_LOG)
        return False

    system = platform.system().lower()

    try:
        if system == "linux":
            return _import_cert_linux(cert_path)
        elif system == "windows":
            return _import_cert_windows(cert_path)
        else:
            util.log_msg(
                f"Unsupported operating system: {system}", mode=util.STDERR_LOG
            )
            return False

    except Exception as e:
        util.log_msg(f"Failed to import SSL certificate: {e}", mode=util.STDERR_LOG)
        return False


def _import_cert_linux(cert_path: pathlib.Path) -> bool:
    """Import certificate on Linux systems."""
    # Try common Linux certificate directories
    cert_dirs = [
        "/usr/local/share/ca-certificates/",
        "/etc/ssl/certs/",
        "/etc/pki/ca-trust/source/anchors/",
    ]

    cert_name = cert_path.name
    imported = False

    for cert_dir in cert_dirs:
        cert_dir_path = pathlib.Path(cert_dir)
        if cert_dir_path.exists() and cert_dir_path.is_dir():
            try:
                target_path = cert_dir_path / cert_name

                # Copy certificate to system directory
                subprocess.run(
                    ["sudo", "cp", str(cert_path), str(target_path)],
                    check=True,
                    capture_output=True,
                    text=True,
                    timeout=30,
                )

                # Update certificate store
                if cert_dir == "/usr/local/share/ca-certificates/":
                    subprocess.run(
                        ["sudo", "update-ca-certificates"],
                        check=True,
                        capture_output=True,
                        text=True,
                        timeout=60,
                    )
                elif cert_dir == "/etc/pki/ca-trust/source/anchors/":
                    subprocess.run(
                        ["sudo", "update-ca-trust"],
                        check=True,
                        capture_output=True,
                        text=True,
                        timeout=60,
                    )

                util.log_msg(
                    f"Certificate imported to {target_path}", mode=util.STDOUT_LOG
                )
                imported = True
                break

            except (subprocess.CalledProcessError, subprocess.TimeoutExpired):
                # Try next directory if this one fails
                continue

    if not imported:
        # Fallback: set environment variable
        util.log_msg(
            "Could not import to system store, using environment variable " "fallback",
            mode=util.STDOUT_LOG,
        )
        # This would typically be set by the calling script
        # os.environ['REQUESTS_CA_BUNDLE'] = str(cert_path)
        # os.environ['SSL_CERT_FILE'] = str(cert_path)

    return True


def _import_cert_windows(cert_path: pathlib.Path) -> bool:
    """Import certificate on Windows systems."""
    try:
        # Use certutil to import the certificate
        subprocess.run(
            ["certutil", "-addstore", "Root", str(cert_path)],
            check=True,
            capture_output=True,
            text=True,
            timeout=60,
        )

        util.log_msg(
            "Certificate imported to Windows certificate store", mode=util.STDOUT_LOG
        )
        return True

    except subprocess.CalledProcessError as e:
        util.log_msg(
            f"Failed to import certificate with certutil: {e.stderr}",
            mode=util.STDERR_LOG,
        )
        return False


def download_and_import_ssl_certificates(servers: List[Tuple[str, int]]) -> bool:
    """Download and import SSL certificates from multiple servers.

    Args:
        servers: List of (hostname, port) tuples

    Returns:
        True if all certificates were processed successfully, False otherwise
    """
    success = True

    for hostname, port in servers:
        try:
            # Download certificate
            cert_path = download_ssl_certificate(hostname, port)

            # Import to system trust store
            if not import_ssl_certificate_to_system(cert_path):
                util.log_msg(
                    f"Failed to import certificate for {hostname}:{port}",
                    mode=util.STDERR_LOG,
                )
                success = False

        except Exception as e:
            util.log_msg(
                f"Error processing {hostname}:{port}: {e}", mode=util.STDERR_LOG
            )
            success = False

    return success



"""
mmm config file.
"""

ALIAS = "mmm50"

# 3m50 thermostat IP addresses (on local net)
# user should configure these zones and IP addresses for their application.
MAIN_3M50 = 0  # zone 0
BASEMENT_3M50 = 1  # zone 1

# all environment variables specific to this thermostat type
env_variables = {}

# min required env variables on all runs
required_env_variables = {}

# supported thermostat configs
supported_configs = {
    "module": "mmm",
    "type": 2,
    "zones": [0, 1],
    "modes": ["OFF_MODE", "HEAT_MODE", "COOL_MODE", "UNKNOWN_MODE"],
    "zip_code": "55760",  # Zip code for outdoor weather data
}

# metadata dict
# 'zone_name' is returned by self.get_zone_name
# 'host_name' is used for dns lookup of IP address for each zone
# 'ip_address' key (if present is used for hard-coding IP address
metadata = {
    MAIN_3M50: {
        "zone_name": "Main Level",
        "host_name": "thermostat-fd-b3-be.lan",
    },
    BASEMENT_3M50: {
        "zone_name": "Basement",
        "host_name": "thermostat-27-67-11.lan",
    },
}


def get_available_zones():
    """
    Return list of available zones.

    for this thermostat type, available zones is all zones.

    inputs:
        None.
    returns:
        (list) available zones.
    """
    return supported_configs["zones"]


default_zone = supported_configs["zones"][0]
default_zone_name = metadata[default_zone]["zone_name"]

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "18",  # poll time in sec
    "358",  # reconnect time in sec
    "2",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
]

# flag to check thermostat response time during basic checkout
check_response_time = False



"""
kumolocal config file.
"""

ALIAS = "kumolocal"

# thermostat zones
MAIN_LEVEL = 0  # zone 0
BASEMENT = 1  # zone 1

# constants
MAX_HEAT_SETPOINT = 68
MIN_COOL_SETPOINT = 70

# all environment variables specific to this thermostat type
env_variables = {
    "KUMO_USERNAME": None,
    "KUMO_PASSWORD": None,
}

# supported thermostat configs
supported_configs = {
    "module": "kumolocal",
    "type": 5,
    "zones": [MAIN_LEVEL, BASEMENT],
    "modes": [
        "OFF_MODE",
        "HEAT_MODE",
        "COOL_MODE",
        "DRY_MODE",
        "AUTO_MODE",
        "UNKNOWN_MODE",
    ],
    "zip_code": "55760",  # Zip code for outdoor weather data
}

# metadata dict
# 'zone_name' is a placeholder, used at Thermostat class level.
# 'zone_name' is updated by device memory via Zone.get_zone_name()
# 'host_name' is used for DNS lookup to determine if device
# 'ip_address' is just for reference.
# 'local_net_available' is set by local network detection
metadata = {
    MAIN_LEVEL: {
        "ip_address": "192.168.86.229",  # local IP, for ref only.
        "zone_name": "Main Level",  # customize for your site.
        "host_name": "tbd",  # used for DNS lookup
        "local_net_available": None,  # updated by local network detection
    },
    BASEMENT: {
        "ip_address": "192.168.86.236",  # local IP, for ref only.
        "zone_name": "Basement",  # customize for your site.
        "host_name": "tbd",  # used for DNS lookup
        "local_net_available": None,  # updated by local network detection
    },
}


def get_available_zones():
    """
    Return list of available zones.

    for this thermostat type, available zones is all zones.

    inputs:
        None.
    returns:
        (list) available zones.
    """
    return supported_configs["zones"]


default_zone = supported_configs["zones"][0]
default_zone_name = metadata[default_zone]["zone_name"]

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "16",  # poll time in sec
    "356",  # reconnect time in sec
    "4",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
]

# flag to check thermostat response time during basic checkout
check_response_time = False
